{-# OPTIONS_GHC -fglasgow-exts #-}

module Transform.Relation.Rel where

import qualified Data.Set as Set
import Data.Set(Set) -- to have the constructor
-- import Data.Relation.SetOfPairs
import Control.Monad
import Data.Maybe
import Data.List (isPrefixOf)
import Control.Monad.RWS
import Prelude hiding (all)
import Transform.Relation.Type 
import Char

import Debug.Trace

--------------------------------------------------------------------------------
-- ABSTRACT SYNTAX
--------------------------------------------------------------------------------

-- | Abstract syntax (GADT)
data R f where
  Rvar   :: String -> [Prop] -> R x

  Rempty :: R (Rel a b)
  Rtop   :: R (Rel a b)
  Rid    :: R (Rel a a)
  Rinv   :: R (Rel a b) -> R (Rel b a)
  Rnot :: R (Rel a b) -> R (Rel a b)
  Rcomp  :: Type b -> R (Rel b c) -> R (Rel a b) -> R (Rel a c)
  
  Rker   :: Type b -> R (Rel a b) -> R (Rel a a)
  Rimg   :: Type a -> R (Rel a b) -> R (Rel b b)

  Rdom   :: Type b -> R (Rel a b) -> R (Rel a a)
  Rrng   :: Type a -> R (Rel a b) -> R (Rel b b)
  
  Rinter :: R (Rel a b) -> R (Rel a b) -> R (Rel a b)
  Runion :: R (Rel a b) -> R (Rel a b) -> R (Rel a b)

  Runder :: Type b -> R (Rel a b) -> R (Rel c b) -> R (Rel c a)
  Rover  :: Type b -> R (Rel b c) -> R (Rel b a) -> R (Rel a c)

  Rincl  :: Type f -> R f -> R f -> R Bool  
  Rand   :: R Bool -> R Bool -> R Bool
  

roverride :: Type (Rel a b) -> R (Rel a b) -> R (Rel a b) -> R (Rel a b)
roverride t@(Rel a b) m n = Runion n (Rcomp a m (Rnot (Rdom b n)))

rupdate :: Type (Rel a b) -> R (Rel b b) -> R (Rel a a) -> R (Rel a b) -> R (Rel a b)
rupdate t@(Rel a b) f phi m = roverride t m (Rcomp b f (Rcomp a m phi))


-- | Enumeration type for taxonomy properties
data Prop = Simple | Entire | Injective | Surjective
          | Coreflexive
          deriving (Show, Eq, Ord)
          
-- Infix operators
(.*.) r s = Rcomp typeof r s
(.=<.) r s = Rincl typeof r s

-- Special relations
bang :: R (Rel a One)
bang = Rvar "!" [Entire,Simple,Surjective]


-- | Showing relational algebra expressions
instance Show (R f) where
  show = rshow'
  
-- | Show with parentheses around composition.
rshow :: R f -> String
rshow Rempty           = "Bottom"
rshow Rtop             = "Top"
rshow Rid              = "Id"
rshow (Rinv r)         = rshow r++"'"
rshow (Rnot r)       = "~("++rshow r++")"
rshow (Rcomp _ r r')   = "("++rshow' r++" . "++rshow' r'++")"
rshow (Rker _ r)       = "Ker("++rshow r++")"
rshow (Rimg _ r)       = "Img("++rshow r++")"
rshow (Rdom t r)       = "Dom("++rshow r++")"
rshow (Rrng _ r)       = "Rng("++rshow r++")"
rshow (Rinter r r')    = "("++rshow r++ " TT " ++rshow r'++")"
rshow (Runion r r')    = "("++rshow r++ " U "++rshow r'++")"
rshow (Runder _ r r')  = "(Runder "++rshow r++" \\ "++rshow r'++")"
rshow (Rover _ r r')   = "(Rover "++rshow r++" / "++rshow r'++")"
rshow (Rincl _ r r')   = rshow' r++" <= "++rshow' r'
rshow (Rand r r')      = "("++rshow' r++" /\\ "++rshow' r'++")"
rshow (Rvar s _ )      = s

-- | Printing composition without parentheses.
rshow' :: R f -> String
rshow' (Rcomp _ r r')  = rshow' r++" . "++rshow' r'
rshow' r               = rshow r 

-- | Equality test
instance Eq (R f) where
  f == g   = req f g
 
req :: R f -> R f -> Bool
req Rempty         Rempty            = True
req Rtop           Rtop              = True
req Rid            Rid               = True
req (Rinv x)       (Rinv x')         = x==x'
req (Rnot x)     (Rnot x')       = x==x'
req (Rcomp t x y)  (Rcomp t' x' y')  = case teq' t t' of
  Just Eq  -> req x x' && req y y'
  _        -> False
req (Rker t x)     (Rker t' x')      = case teq' t t' of
  Just Eq  -> x==x'
  _        -> False
req (Rimg t x)     (Rimg t' x')      = case teq' t t' of
  Just Eq  -> x==x'
  _        -> False
req (Rdom t x)     (Rdom t' x')      = case teq' t t' of
  Just Eq  -> x==x'
  _        -> False
req (Rrng t x)     (Rrng t' x')      = case teq' t t' of
  Just Eq  -> x==x'
  _        -> False
req (Rinter x y)  (Rinter x' y')     = req x x' && req y y'
req (Runion x y)  (Runion x' y')     = req x x' && req y y'
req (Runder t x y) (Runder t' x' y') = case teq' t t' of
  Just Eq  -> req x x' && req y y'
  _        -> False
req (Rover t x y)  (Rover t' x' y')  = case teq' t t' of
  Just Eq  -> req x x' && req y y'
  _        -> False
req (Rincl f x y) (Rincl f' x' y') = case teq' f f' of
  Just Eq  ->  req x x' && req y y'
  _        -> False
req (Rand x y) (Rand x' y') = req x x' && req y y'
req (Rvar s p) (Rvar t p')           = 
  s==t && (Set.fromList p) == (Set.fromList p')
req _            _              = False


--------------------------------------------------------------------------------
-- Individual REWRITE RULES for relation operations 
-- LEVEL 1 - (only symbols)
--------------------------------------------------------------------------------

-- | Composition, Inclusion AXIOMS
-- Right associativity of composition:  (r . s) . t = r . (s . t)
comp_assocr :: Rule
comp_assocr _ (Rcomp a (Rcomp b r s) t)    = return (Rcomp b r (Rcomp a s t))
comp_assocr _ _                            = mzero

-- Left associativity of composition:  t . (r . s) = (t . r) . s
comp_assocl :: Rule
comp_assocl _ (Rcomp a t (Rcomp b r s))    = return (Rcomp b (Rcomp a t r) s)
comp_assocl _ _                            = mzero

-- plus identity: r = Id . r
comp_idintro :: Rule
comp_idintro (Rel a b)  r          = return (Rcomp b Rid r)
comp_idintro _ _                   = mzero

-- identity: Id . r = r . Id = r
comp_id :: Rule
comp_id t (Rcomp _ Rid r)          = success "comp_id/2" t r
comp_id t (Rcomp _ r Rid)          = success "comp_id/2" t r
comp_id _ _                        = mzero

-- composing with bottom: _L . r = r . _L = _L
comp_empty :: Rule
comp_empty t (Rcomp _ Rempty r)    = success "comp_empty/1" t Rempty
comp_empty t (Rcomp _ r Rempty)    = success "comp_empty/2" t Rempty
comp_empty _ _                     = mzero

-- ~Id = Empty
neg_id :: Rule
neg_id  t (Rnot Rid)     = success "neg_id" t Rempty
neg_id _ _                         = mzero
    
-- (~ r . r ), (r . ~ r), = Empty
neg_comp :: Rule
neg_comp = neg_comp1 ||| neg_comp2
  where
    neg_comp1, neg_comp2 :: Rule
    neg_comp1 t@(Rel a c) (Rcomp b (Rnot r) r') = do
      Eq <- teq a b
      Eq <- teq a c
      guard (r==r')
      success "neg_comp/1" t Rempty
    neg_comp1 t@(Rel a c) (Rcomp b r (Rnot r')) = do
      Eq <- teq a b
      Eq <- teq a c
      success "neg_comp/2" t Rempty
    neg_comp1 _ _ = mzero
    neg_comp2 t@(Rel a d) (Rcomp b (Rcomp c h (Rnot r)) r') = do
      Eq <- teq a b
      Eq <- teq a c
      guard (r==r')
      success "neg_comp/3" t Rempty
    neg_comp2 t@(Rel a d) (Rcomp b (Rcomp c h r) (Rnot r')) = do
      Eq <- teq a b
      Eq <- teq a c
      guard (r==r')
      success "neg_comp/4" t Rempty
    neg_comp2 _ _ = mzero

-- r <= r = True
incl_refl :: Rule
incl_refl t e@(Rincl _ r r')
  | r==r'
  = success "incl_refl " t (Rvar "True" [])
incl_refl _ _                      = mzero

-- _L <= r = True
incl_bottom :: Rule
incl_bottom t (Rincl _ Rempty r')
  = success "incl_bottom" t (Rvar "True" [])
incl_bottom _ _                      = mzero

-- r <= T = True
incl_top :: Rule
incl_top t (Rincl _ r' Rtop)
  = success "incl_top" t (Rvar "True" [])
incl_top _ _                      = mzero
    
true :: R Bool
true = (Rvar "True" [])

-- empty converse: _L ' = _L
inv_empty :: Rule
inv_empty t (Rinv Rempty)= success "inv_empty" t Rempty
inv_empty _ _          = mzero

-- identity converse: Id' = Id
inv_id :: Rule
inv_id t (Rinv Rid) = success "inv_id" t Rid 
inv_id _ _          = mzero

-- top converse: T' = T
inv_top :: Rule
inv_top _ (Rinv Rtop) = return Rtop
inv_top _ _           = mzero

-- converse involution: r'' = r
inv_inv :: Rule
inv_inv t e@(Rinv (Rinv r))   = success "inv_inv" t r
inv_inv _ _                 = mzero

-- converse contravariance: (r . s)' = s' . r'
inv_comp :: Rule
inv_comp tp (Rinv (Rcomp t r r'))   = success "inv_comp" tp (Rcomp t (Rinv r') (Rinv r))
inv_comp _ _                       = mzero

-- dom of a converse: dom(r')= rng r
dom_inv :: Rule
dom_inv tp (Rdom t (Rinv r))   = success "dom_inv" tp (Rrng t r)
dom_inv _ _                 = mzero

-- rng of converse: rng(r') = dom r
rng_inv :: Rule
rng_inv tp (Rrng t (Rinv r)) = success "rng_inv" tp (Rdom t r) 
rng_inv _ _                 = mzero

{- Not yet
-- | DEDEKIND VARIATIONS
-- (left modular)
-- (right modular)
-- (weak distr.)
-}

-- | Intersection LAWS
-- Inter. associativity: (r TT s) TT t = r TT (s TT t)
inter_assoc :: Rule
inter_assoc _ (Rinter (Rinter r s) t) = return (Rinter r (Rinter s t))
inter_assoc _ _                       = mzero

-- Inter. commutativity: r TT s = s TT r
inter_comm :: Rule
inter_comm _ (Rinter r s) = return (Rinter s r)
inter_comm _ _            = mzero

-- Inter. idempotence: r TT r = r
inter_idem :: Rule
inter_idem t (Rinter r s) 
  | r==s                  = success "inter_idem" t r
inter_idem _ _            = mzero

inter_idem' :: Rule
inter_idem' (Rel a c) (Rinter r r')  = do
  Eq <- teq a c
  guard (r==r')
  return r
inter_idem' _ _                      = mzero

-- (r TT s) . t = (r . t) TT (s . t)
inter_fusion :: Rule
inter_fusion x (Rcomp y (Rinter r s) t)
  = success "inter_right_fusion" x (Rinter (Rcomp y r t) (Rcomp y s t))
inter_fusion x (Rcomp y t (Rinter r s))
  = success "inter_left_fusion" x (Rinter (Rcomp y t r) (Rcomp y t s))
inter_fusion _ _ = mzero

-- z <= (x TT y) = z <= x /\ z <= y
inter_split :: Rule
inter_split t (Rincl f z (Rinter x y))
  = success "inter_split" t (Rand (Rincl f z x) (Rincl f z y))
inter_split _ _ = mzero

-- | Union LAWS
-- (r U s)' = r' U s'
union_contravariance :: Rule
union_contravariance t (Rinv (Runion r s))
  = success "union_contravariance" t (Runion (Rinv r) (Rinv s))
union_contravariance _ _ = mzero

-- (r U s) . t = (r . t) U (s . t)
union_fusion :: Rule
union_fusion x (Rcomp y (Runion r s) t)
  = success "union_right_fusion" x (Runion (Rcomp y r t) (Rcomp y s t))
union_fusion x (Rcomp y t (Runion r s))
  = success "union_left_fusion" x (Runion (Rcomp y t r) (Rcomp y t s))
union_fusion _ _ = mzero

-- (x U y) <= z = x <= z /\ y <= z
union_split :: Rule
union_split t (Rincl f (Runion x y) z)
  = success "union_split" t (Rand (Rincl f x z) (Rincl f y z))
union_split _ _ = mzero

-- | Kernel LAWS
-- definition: ker r = r' . r
ker_def :: Rule
ker_def t e@(Rker b r)  = success "ker_def" t (Rcomp b (Rinv r) r)
ker_def _ _             = mzero

ker_def' :: Rule
ker_def' t@(Rel a c) (Rcomp b (Rinv r) r')  = do
  Eq <- teq a c
  guard (r==r')
  success "ker_def_rev" t (Rker b r)
ker_def' _ _                              = mzero

-- duality: ker (r') = Img r
ker_dual :: Rule
ker_dual _ (Rker b (Rinv r))    = return (Rimg b r)
ker_dual _ _                    = mzero

-- symmetry: (ker r)' = ker r
ker_symm :: Rule
ker_symm _ (Rinv (Rker b r))    = return (Rker b r)
ker_symm _ _                    = mzero

-- kernel composition: ker (r . s) = s' . (ker r . s)
ker_comp :: Rule
ker_comp _ (Rker b (Rcomp c r s)) = 
           return (Rcomp c (Rinv s) (Rcomp c (Rker b r) s))
ker_comp _ _                    = mzero

{- Not yet
-- (monotonicity)
-- (intro)
-- (inter. kernel)
-}

-- | Image Laws
-- definition: Img r = r . r'
img_def :: Rule
img_def t e@(Rimg a r)  = success "img_def" t (Rcomp a r (Rinv r))
img_def _ _             = mzero

img_def' :: Rule
img_def' t@(Rel a c) (Rcomp b r (Rinv r'))  = do
  Eq <- teq a c
  guard (r==r')
  success "img_def_rev" t (Rimg b r)
img_def' _ _                              = mzero

-- duality: Img (r') = ker r
img_dual :: Rule
img_dual _ (Rimg a (Rinv r))    = return (Rker a r)
img_dual _ _                    = mzero

-- symmetry: (Img r)' = Img r
img_symm :: Rule
img_symm _ (Rinv (Rimg a r))    = return (Rimg a r)
img_symm _ _                    = mzero

-- intro: (Img r) . r = r
img_intro :: Rule
img_intro (Rel a b) (Rcomp b' (Rimg a' r) r')   =do
  Eq <- teq a b
  Eq <- teq a a'  
  Eq <- teq b b'  
  guard (r==r')
  return  r 
img_intro _ _                        = mzero

-- | Order Taxonomy
-- connected: r U r' = T
ord_connect :: Rule
ord_connect (Rel a b) (Runion r (Rinv r'))  = do
  Eq <- teq a b
  guard (r==r')
  return Rtop
ord_connect _ _                     = mzero

-- Not yet
-- (order reflexivity)
-- (order transitivity)
-- (order anti-symmetry)
-- (monotonicity)

-- | Converse LAWS
-- (reduction)
-- (wrap)
-- (dist over inter.)

-- | MISCELLANEOUS
-- (Dedekind: T=id,R=Rc)
-- (Dedekind: S=id)


-- | Equality AXIOMS
-- (ping-pong)
-- (indirection)
-- | Converse, intersection AXIOMS
-- (universal prop. of intersection)
-- (converse monotonicity)
-- (left modular)
-- (reflexive)
-- (coreflexive)
-- (transitive)
-- (symmetric)
-- (anti-symmetric)
-- (entire)
-- (simple)
-- (surjection)
-- (injection)

-- | Correflexives
-- (symm & transitive)
-- (cancellation)
-- (distributivity)

-- | Bang Rules
-- | Top Rules
-- | Bottom Rules
-- | Unit Rules
-- | Zero Rules


-- | Galois Connections
-- r . x <= y == x <= r\y
gc_rightd :: Rule
gc_rightd _ (Rincl (Rel a c) (Rcomp b r x) y)
  = return (Rincl (Rel a b) x (Runder c r y))
gc_rightd _ _ = mzero

-- x . r <= y == x <= y/r
gc_leftd :: Rule
gc_leftd _ (Rincl (Rel b c) (Rcomp a x r) y)
  = return (Rincl (Rel a c) x (Rover b y r))
gc_leftd _ _ = mzero

-- | Almost Galois Connections
-- dom x <= y == x <= x . y
gc_VRdom :: Rule
gc_VRdom _ (Rincl (Rel a a') (Rdom b x) y) = do
  Eq <- teq a a'
  return (Rincl (Rel a b) x (Rcomp a x y))
gc_VRdom _ _ = mzero

-- Rng x <= y == x <= y . x
gc_VRrng :: Rule
gc_VRrng _ (Rincl (Rel b b') (Rrng a x) y) = do
  Eq <- teq b b'
  return (Rincl (Rel a b) x (Rcomp b y x))
gc_VRrng _ _ = mzero

--------------------------------------------------------------------------------
-- Individual REWRITE RULES for relation operations 
-- LEVEL 2 - (symbols + taxonomy)
--------------------------------------------------------------------------------

-- | Galois Connections
-- f . x <= y  == x <= f' y 
gc_shunt :: Rule
gc_shunt t (Rincl (Rel a c) (Rcomp b f x) y)
  | isEntire (Rel b c) f && isSimple (Rel b c) f
  = success "gc_shunt" t (Rincl (Rel a b) x (Rcomp c (Rinv f) y))
gc_shunt _ _ = mzero

-- x . f' <= y == x <= y . f
gc_shuntc :: Rule
gc_shuntc t (Rincl (Rel a c) (Rcomp b x (Rinv f)) y)
  | isEntire (Rel b a) f && isSimple (Rel b a) f
  = success "gc_shuntc" t (Rincl (Rel b c) x (Rcomp a y f))
gc_shuntc _ _ = mzero

-- | Almost Galois Connections
-- 
gc_RSshunt' :: Rule
gc_RSshunt' = innermost comp_assocr >>> gc_RSshunt >>> innermost comp_assocl

-- r . x <= y (r simple) == dom r . x <= r' . y
gc_RSshunt :: Rule         
gc_RSshunt t (Rincl (Rel a c) (Rcomp b rs x) y)
  | isSimple (Rel b c) rs 
  = success "gc_RSshunt" t 
            (Rincl (Rel a b) (Rcomp b (Rdom c rs) x) (Rcomp c (Rinv rs) y))
gc_RSshunt _ _ = mzero

-- x . r <= y (r simple) == x . dom r <= y . r 
gc_RSshuntc :: Rule
gc_RSshuntc t (Rincl (Rel a c) (Rcomp b x (Rinv rs)) y)
  | isSimple (Rel b a) rs
  = success "gc_RSshuntc" t 
            (Rincl (Rel b c) (Rcomp b x (Rdom a rs)) (Rcomp a y rs))
gc_RSshuntc _ _ = mzero

{-gc_RSshunt Bool (Rincl (Rel a c) x y) = do undefined
gc_RSshunt Bool _ = do undefined

  where
    aux :: Type b -> Type (Rel z c) -> R (Rel z c) -> Maybe (R (Rel z b))
    aux b (Rel z c) (Rcomp b' r s) = do
      r' <- aux b' (Rel b' c) r
      return (Rcomp b' r' s)
    aux b (Rel z c) r = do
      Eq <- teq b z
      guard (isSimple (Rel b c) r)
      return (Rdom c r)


  result <- aux y (Rel a a) Rid (Rel a c) x 
  success "gc_RSshunt" t result
 where
  aux :: R (Rel a c) -> Type (Rel a .) -> R (Rel a .) -> Rule
  aux y (Rel a z) s' (Rel x c) (Rcomp b r s) = aux (Rel a .) (Rcomp z s s') (Rel b c) r
  aux y (Rel a b') x (Rel b c) r = do
    Eq <- teq b b'
    guard (isSimple (Rel b c) r)
    success "gc_RSshunt" Bool 
            (Rincl (Rel a b) (Rcomp b (Rdom b r) x) 
                             (Rcomp c (Rinv r) y))
-}   

-- | Identity elimination (f'.f con f injectiva y entire)
ker_elim:: Rule
ker_elim t@(Rel a b) (Rcomp c (Rinv r) r')    = do
  Eq <- teq a b
  guard (r==r')
  guard( isEntire (Rel a c) r'  && isInjective (Rel a c) r') 
  success "ker_elim/1" t Rid
ker_elim t@(Rel a d) (Rcomp c (Rcomp b h (Rinv r)) r')    = do
  Eq <- teq a b
  guard (r==r')
  guard( isEntire (Rel a c) r'  && isInjective (Rel a c) r') 
  success "ker_elim/2" t  h
ker_elim x y 
  = mzero
  
-- | Kernel of constant function is top
--   (const r)' . const r = top
ker_elim_constant :: Rule
-- Sea r entera, simple, constant:  r' . r = top 
ker_elim_constant t@(Rel a b) (Rcomp c (Rinv r) r')    = do
  Eq <- teq a b
  guard (r==r')
  guard( isEntire (Rel a c) r'  && isSimple (Rel a c) r'  && isConstant r')
  success "ker_elim_constant/1" t Rtop
-- Sea r entera, simple, constant:  (h . r') . r = h . top  
ker_elim_constant t@(Rel a d) (Rcomp c (Rcomp b h (Rinv r)) r')    = do
  Eq <- teq a b
  guard (r==r')
  guard( isEntire (Rel a c) r'  && isSimple (Rel a c) r'  && isConstant r')
  success "ker_elim_constant/2" t  (Rcomp b h Rtop)
ker_elim_constant x y 
  = mzero

-- | Identity elimination (R.R' with R surjective and simple)
img_elim:: Rule
img_elim = img_elim1 ||| img_elim2
  where
    img_elim1, img_elim2 :: Rule
    img_elim1 t@(Rel a b) (Rcomp c r' (Rinv r))    = do
      Eq <- teq a b
      guard (r==r')
      guard( isSimple (Rel c a) r'  && isSurjective (Rel c a) r') 
      success "img_elim/1" t Rid
    img_elim1 x y 
      = mzero
    img_elim2 t@(Rel a d) (Rcomp c (Rcomp b h r') (Rinv r))    = do
      Eq <- teq a b
      guard (r==r')
      guard( isSimple (Rel c a) r'  && isSurjective (Rel c a) r') 
      success "img_elim/2" t h
    img_elim2 x y 
      = mzero
  
-- | Domain elimination (R . dom R = R)
dom_elim :: Rule
dom_elim = dom_elim1 ||| dom_elim2
  where
    dom_elim1, dom_elim2 :: Rule
    dom_elim1 t@(Rel a b) e@(Rcomp c r (Rdom d r')) = do
      Eq <- teq b d
      guard (r==r')
      success "dom_elim/1" t r
    dom_elim1 _ _ = mzero
    dom_elim2 t@(Rel a z) e@(Rcomp c (Rcomp b h r) (Rdom d r')) = do
      Eq <- teq b d
      guard (r==r')
      success "dom_elim/2" t (Rcomp b h r)
    dom_elim2 _ _ = mzero

-- | Range elimination (rng R . R = R)
rng_elim :: Rule
rng_elim =  rng_elim1 ||| rng_elim2
  where
    rng_elim1, rng_elim2 :: Rule
    rng_elim1 t@(Rel a b) (Rcomp c (Rrng d r) r') = do
      Eq <- teq a d
      guard (r==r')
      success "rng_elim/1" t r
    rng_elim1 _ _ = mzero
    rng_elim2 t@(Rel a z) (Rcomp c (Rcomp b h (Rrng d r)) r') = do
      Eq <- teq a d
      guard (r==r')
      success "rng_elim/2" t (Rcomp b h r)
    rng_elim2 _ _ = mzero
  
-- | Domain cancels against relation itself.
--   R . phi . dom R = R . phi
dom_cancel :: Rule
dom_cancel t@(Rel a b) (Rcomp a' x (Rdom c r)) =
    case aux (Rel a c) r (Rel a' b) x of
      Just msg  -> success msg t x
      Nothing   -> mzero
  where
    aux (Rel a c) r (Rel a' b) x = do
      (aux1 (Rel a c) r (Rel a' b) x) `mplus` 
       (aux2 (Rel a c) r (Rel a' b) x) `mplus` 
        (aux3 (Rel a c) r (Rel a' b) x)
    aux1,aux2,aux3 :: Type (Rel a b) -> R (Rel a b) -> (forall f . Type f -> R f -> Maybe String)
    aux1 t@(Rel a b) r t'@(Rel a' b') r' = do 
      Eq <- teq t t'
      guard (r==r')
      return "dom_cancel/1"
    aux2 (Rel a b) r (Rel a' c') (Rcomp b' x r') = do 
      Eq <- teq (Rel a' b') (Rel a b)
      guard (r==r')
      return "dom_cancel/2"
    aux2 _ _ _ _ = mzero
    aux3 t@(Rel a b) r (Rel a' c') (Rcomp b' x r') = do
      Eq <- teq a' b'
      guard (isCoreflexive a' r')
      aux t r (Rel b' c') x
    aux3 _ _ _ _ = mzero
dom_cancel _ _ = mzero

    
not_dom_cancel :: Rule
not_dom_cancel t@(Rel a b) (Rcomp a' x (Rnot (Rdom c r))) =
    case aux (Rel a c) r (Rel a' b) x of
      Just msg  -> success msg t Rempty
      Nothing   -> mzero
  where
    aux (Rel a c) r (Rel a' b) x = do
      (aux1 (Rel a c) r (Rel a' b) x) `mplus` 
       (aux2 (Rel a c) r (Rel a' b) x) `mplus` 
        (aux3 (Rel a c) r (Rel a' b) x)
    aux1,aux2,aux3 :: Type (Rel a b) -> R (Rel a b) -> (forall f . Type f -> R f -> Maybe String)
    aux1 t@(Rel a b) r t'@(Rel a' b') r' = do 
      Eq <- teq t t'
      guard (r==r')
      return "not_dom_cancel/1"
    aux2 (Rel a b) r (Rel a' c') (Rcomp b' x r') = do 
      Eq <- teq (Rel a' b') (Rel a b)
      guard (r==r')
      return "not_dom_cancel/2"
    aux2 _ _ _ _ = mzero
    aux3 t@(Rel a b) r (Rel a' c') (Rcomp b' x r') = do
      Eq <- teq a' b'
      guard (isCoreflexive a' r')
      aux t r (Rel b' c') x
    aux3 _ _ _ _ = mzero
not_dom_cancel _ _ = mzero

-- dom (f . M) = dom M
dom_comp_cancel :: Rule
dom_comp_cancel t (Rdom c (Rcomp b f m))
  | isEntire (Rel b c) f && isSimple (Rel b c) f
  = success "dom_comp_cancel" t (Rdom b m)
dom_comp_cancel _ _ = mzero

--------------------------------------------------------------------------------
-- Relational Taxonomy
--------------------------------------------------------------------------------

-- | Helper functions to create propositions about (co)reflexivity.
propCoreflexive t r             =  Rincl (Rel t t) r Rid
propReflexive t r               =  Rincl (Rel t t) Rid r

-- | Helper functions to create propositions about basic properties.
propSimple (Rel a b) r          = propCoreflexive b (Rimg a r)
propSurjective (Rel a b) r      = propReflexive b (Rimg a r)
propInjective (Rel a b) r       = propCoreflexive a (Rker b r)
propEntire (Rel a b) r           = propReflexive a (Rker b r)

-- | Test whether relation is simple.
isSimple:: Type (Rel a b) -> R (Rel a b) -> Bool
isSimple _ Rid = True
isSimple _ (Rvar _ lp) = Set.member Simple (Set.fromList lp)
isSimple (Rel a b) (Rinv r) = isInjective (Rel b a) r
isSimple (Rel a c) (Rcomp b s r) 
  = isSimple (Rel a b) r && isSimple (Rel b c) s
isSimple _ _ = False
{- This loops!!
isSimple t r = case check (propSimple t r) of
  Just (Rvar "True" [])   -> True
  _                       -> False
-}


-- | Test whether relation is entire.
isEntire :: Type (Rel a b) -> R (Rel a b) -> Bool
isEntire _ (Rvar _ lp) = Set.member Entire (Set.fromList lp)
isEntire (Rel a b) (Rinv r) = isSurjective (Rel b a) r
isEntire (Rel a _) (Rdom b r) = isEntire (Rel a b) r
--is Entire (Rel a b) r = isReflexive (ker r)
isEntire (Rel a b) r = False

-- | Test whether relation is injective.
isInjective :: Type (Rel a b) -> R (Rel a b) -> Bool
isInjective _ (Rvar _ lp) = Set.member Injective (Set.fromList lp)
--isInjective (Rel a b) r = isCoreflexive (ker r)
isInjective (Rel a b) r = False

-- | Test whether relation is surjective.
isSurjective :: Type (Rel a b) -> R (Rel a b) -> Bool
isSurjective _ (Rvar _ lp) = Set.member Surjective (Set.fromList lp)
--isSurjective (Rel a b) r = isReflexive (img r)
isSurjective (Rel a b) r = False

-- | Test whether relation is a representation.
isRepresentation :: Type (Rel a b) -> R (Rel a b) -> Bool
isRepresentation t r = isInjective t r && isEntire t r 

-- | Test whether relation is functional.
isFunction:: Type (Rel a b) -> R (Rel a b) -> Bool
isFunction  _ (Rvar _ lp) = (Set.member Simple slp) && (Set.member Entire slp) 
   where slp = Set.fromList lp
isFunction _ _ = False
{-
isFunction  (Rel a b) r = x == Rid
          where (Just x) = simplify (Rel a a) (Rcomp b (Rinv r) r )
-}

-- | Test whether relation is an abstraction.
isAbstraction :: Type (Rel a b) -> R (Rel a b) -> Bool
isAbstraction t r = isSimple t r && isSurjective t r

-- | Test whether relation is an injective function.
isInjection :: Type (Rel a b) -> R (Rel a b) -> Bool
isInjection t r = isRepresentation t r && isFunction t r

-- | Test whether relation is a surjective function.
isSurjection :: Type (Rel a b) -> R (Rel a b) -> Bool
isSurjection t r = isFunction t r && isAbstraction t r

-- | Test whether relation is an isomorphism.
isBijection :: Type (Rel a b) -> R (Rel a b) -> Bool
isBijection t r = isInjection t r && isSurjection t r

-- | ORDERS (Endo-relations)

{--
-- | Test whether relation is reflexive.
isReflexive :: Type (Rel a a) -> R (Rel a a) -> Bool
isReflexive _ (Rvar _ lp) = Rid .<=. r
  where id = identityRel (ent r)

r .<=. id
  where id = identityRel (ent r)
-}

-- | Test whether relation is coreflexive.
isCoreflexive' :: Type a -> Type b -> R (Rel a b) -> Bool
isCoreflexive' a b r = maybe False id $ do
  Eq <- teq a b
  return (isCoreflexive a r)

isCoreflexive :: Type a -> R (Rel a a) -> Bool
isCoreflexive _ (Rvar _ lp)    = Set.member Coreflexive (Set.fromList lp)
isCoreflexive _ (Rdom _ _)     = True
isCoreflexive _ (Rrng _ _)     = True
isCoreflexive (Rel a b) (Rinv r) = isCoreflexive (Rel a b) r
isCoreflexive _ (Rnot _)       = True
isCoreflexive _ (Rimg _ _)     = True
isCoreflexive _ (Rker _ _)     = True
{-
isCoreflexive a (Rcomp b r s)  = case teq a b of
  Just Eq    -> isCoreflexive a r && isCoreflexive a s
  _          -> False 
-}
isCoreflexive _ _ = False

-- | Domain of an entire relation is the identity
dom_id_entire :: Rule
dom_id_entire t@(Rel a _) (Rdom b r) 
  | isEntire (Rel a b) r 
  = success "dom_id_entire/1" t Rid
dom_id_entire t@(Rel z _) (Rcomp a h (Rdom b r)) 
  | isEntire (Rel a b) r 
  = success "dom_id_entire/2" t h
dom_id_entire _ _ = mzero

--------------------------------------------------------------------------------
--  COREFLEXIVES  
--------------------------------------------------------------------------------

corefl_symm :: Rule
corefl_symm t@(Rel a b) e@(Rinv r) = do
  Eq <- teq a b
  guard (isCoreflexive a r)
  success "corefl_symm" t r
corefl_symm _ _ = mzero

corefl_trans :: Rule
corefl_trans = corefl_trans1 ||| corefl_trans2
 where
  corefl_trans1,corefl_trans2 :: Rule
  corefl_trans1 t@(Rel a b) e@(Rcomp c r r') = do
    Eq <- teq a b
    Eq <- teq a c
    guard (isCoreflexive a r)
    guard (r==r')
    success "corefl_trans/1" t r
  corefl_trans1 _ _ = mzero
  corefl_trans2 t@(Rel a d) e@(Rcomp c (Rcomp b h r) r') = do
    Eq <- teq a b
    Eq <- teq a c
    guard (isCoreflexive a r)
    guard (r==r')
    success "corefl_trans/2" t (Rcomp b h r)
  corefl_trans2 _ _ = mzero

{-
-- | Tricky! Performs strengthening of the proof obligation!
corefl_cancelL :: Rule
corefl_cancelL _ e@(Rincl (Rel a b) (Rcomp c r s) r') = do
  Eq <- teq a c
  guard (isCoreflexive a s)
  success ("corefl_cancelL "++show e) (Rincl (Rel a b) r r')
corefl_cancelL _ _ = mzero

-- | Tricky! Performs strengthening of the proof obligation!
corefl_cancelR :: Rule
corefl_cancelR _ e@(Rincl (Rel a b) (Rcomp c s r) r') = do
  Eq <- teq b c
  guard (isCoreflexive b s)
  success ("corefl_cancelR "++show e) (Rincl (Rel a b) r r')
corefl_cancelR _ _ = mzero
-}

-- | Perform argument rule once on any of the direct arguments
--   of an n-ary composition.
once_comp :: Rule -> Rule
once_comp r (Rel a c) (Rcomp b x y) = 
     (do x' <- once_comp r (Rel b c) x; return (Rcomp b x' y))
  `mplus`
     (do y' <- once_comp r (Rel a b) y; return (Rcomp b x y'))
once_comp r t x = r t x 
  
-- | Tricky! Performs strengthening of the proof obligation!
corefl_cancel' :: Rule
corefl_cancel' tp e@(Rincl t r s) = do
    r' <- once_comp corefl2id t r
    implies "monotonicity" tp (Rincl t r' s)
  where
    corefl2id :: Rule
    corelf2id _ Rid = mzero
    corefl2id (Rel a b) r = do
      Eq <- teq a b
      guard (isCoreflexive a r)
      return Rid
    corefl2id _ _ = mzero
corefl_cancel' _ _ = mzero

-- | Perform argument rule once on any of the direct arguments
--   of an n-ary conjuction.
once_and :: Rule -> Rule
once_and r _ (Rand x y) = 
     (do x' <- once_and r Bool x; return (Rand x' y))
  `mplus`
     (do y' <- once_and r Bool y; return (Rand x y'))
once_and r t x = r t x 

all_and :: Rule -> Rule
all_and r t e@(Rand x y) = all (all_and r) t e
all_and r t x = r t x 

and_true :: Rule
and_true t (Rand x (Rvar "True" [])) = success "and/1" t x
and_true t (Rand (Rvar "True" []) y) = success "and/2" t y
and_true _ _ = mzero


--------------------------------------------------------------------------------
--  Generalizations of point-free program algebra  
--------------------------------------------------------------------------------

-- | The bang relation absorbs any entire relation with which it 
--   is composed. We implement the more general rule:
--
--      ! . s  ==> ! . dom s 
--
--   for any s. Together with dom_id_entire (dom r => id, for r entire),
--   this has the same effect.
bang_fusion :: Rule
bang_fusion _ (Rcomp _ _ (Rdom _ _)) = mzero
bang_fusion t@(Rel a One) (Rcomp b s r)
  | isEntire (Rel b One) s && (not (isCoreflexive' a b r))
  = success "bang_fusion" t (Rcomp a bang (Rdom b r))
bang_fusion _ _ = mzero

-- | a more general rule that bang_fusion is valid:
--
--      const k . R = const k . dom R
-- previous version
const_fusion :: Rule
--const_fusion _ (Rcomp _ _ (Rdom _ _)) = mzero
const_fusion t@(Rel a c) (Rcomp b s r)
  | isEntire (Rel b c) s && isConstant s && (not (isCoreflexive' a b r))
--  = success "const_fusion" t (Rcomp a (kk c ("("++show s++")")) (Rdom b r))
  = success "const_fusion" t (Rcomp a (kk c (show s)) (Rdom b r))
const_fusion _ _ = mzero

{-
-- (is applied too: f.R = f.dom R if f is Entire !!!)
const_fusion :: Rule
const_fusion t@(Rel a c) (Rcomp b s@(Rvar n p) r)
  | isEntire (Rel b c) s && (not (isCoreflexive' a b r))
  = success "const_fusion" t (Rcomp a (Rvar n p) (Rdom b r))
const_fusion _ _ = mzero
-}



--------------------------------------------------------------------------------
--  REWRITING  
--------------------------------------------------------------------------------

-- The basic type of rules
type Rule = forall f . Type f -> R f -> Rewrite (R f)

-- The rewrite monad
--   Reader: -- not used at the moment --
--   State: current context
--   Writer: log of rules and intermediate results
-- type Rewrite = RWST DynR [(Ordering,String,String)] Context Maybe
type Rewrite = RWST DynR [(Ordering,String,String)] Context []


-- The context of a term is modeled as a function that takes
-- the current term into the root term. The terms are 
-- dynamically typed. Traversal combinators have the 
-- responsibility of locally extending the context when
-- traversing downward.
type Context = DynR -> DynR

-- Dynamically typed terms of the relational algebra.
data DynR where
  DynR :: Type r -> R r -> DynR

-- Successful application of a rule.
success :: String -> Type a -> R a -> Rewrite (R a)
success n t x = do 
  cxt <- get
  DynR _ r <- return (cxt (DynR t x)) 
  tell [(EQ,n, show r)]
  return x
  
implies :: String -> Type a -> R a -> Rewrite (R a)
implies n t x = do 
  cxt <- get
  DynR _ r <- return (cxt (DynR t x)) 
  tell [(LT,n, show r)]
  return x

-- Application of a rule to a term (monad maybe).
run' :: Rule -> Type a -> R a -> R a
run' s t e = head $ do
  (x,l) <- evalRWST (s t e) (DynR t e) id
  return x

run'' :: Rule -> Type a -> R a -> [R a]
run'' s t e = do
  (x,l) <- evalRWST (s t e) (DynR t e) id
  return x
  
-- Application of a rule to a term, producing the proof trace on stdout 
-- (monad maybe).
run :: Bool -> Rule -> Type a -> R a -> IO ()
run b s t e = do
   (x,l):_ <- return $ evalRWST (s t e) (DynR t e) id
   putStrLn ("  "++show e)
   when b $ sequence_ (map aux l)
 where 
   aux (op,n,y) = putStrLn (ord2op op++"   { "++n++" }")  >> putStrLn ("  "++y)
   ord2op EQ = "=="
   ord2op GT = "=>"
   ord2op LT = "<="

runAt :: Int -> Bool -> Rule -> Type a -> R a -> IO ()
runAt n b s t e = do
   results <- return $ evalRWST (s t e) (DynR t e) id
   let (x,l) = results !! n
   putStrLn ("  "++show e)
   when b $ sequence_ (map aux l)
   unless b $ last (map aux l)
 where 
   aux (op,n,y) = putStrLn (ord2op op++"   { "++n++" }")  >> putStrLn ("  "++y)
   ord2op EQ = "=="
   ord2op GT = "=>"
   ord2op LT = "<="

tf :: Rule -> Rule
tf r Bool p = truthfirst' (r Bool p)
tf r t    e = r t e
 
truthfirst' :: Rewrite (R Bool) -> Rewrite (R Bool)
truthfirst' m = RWST (\r s -> truthfirst (runRWST m r s))
 
truthfirst :: [(R Bool, p, q)] -> [(R Bool, p, q)]
truthfirst =
  uncurry (++) . split (filter (isTrue . fst3)) (filter (not . isTrue . fst3))
  where
    isTrue :: R Bool -> Bool
    isTrue (Rvar "True" _)  = True
    isTrue _                = False

fst3 (a,_,_) = a

split :: (a -> b) -> (a -> c) -> a -> (b,c)
split f g x = (f x, g x)

--------------------------------------------------------------------------------
--  BASIC STRATEGY COMBINATORS  
--------------------------------------------------------------------------------

-- Identity
nop :: Rule
nop _ = return

-- Zero
failure :: Rule
failure _ _ = mzero

-- Sequential composition
(>>>) :: Rule -> Rule -> Rule
(f >>> g) t x = f t x >>= g t


-- left-biased  (Choice for monadic transformations?)
(|+|) :: Rule -> Rule -> Rule
(f |+| g) t x = f t x `mplus` g t x

(|||) :: Rule -> Rule -> Rule
(f ||| g) t x = f t x `mcatch` g t x

class (Monad m) => MonadCatch m where
  mcatch :: m a -> m a -> m a

instance MonadCatch [] where
  -- mcatch :: [a] -> [a] -> [a]
  mcatch [] ys  = ys
  mcatch xs _   = xs

instance (Monoid w, MonadCatch m) => MonadCatch (RWST r w s m) where
 	m `mcatch` n = RWST $ \r s -> runRWST m r s `mcatch` runRWST n r s



{-
-- left-biased choice with PARTIAL backtracking
(|||) :: Rule -> Rule -> Rule
(f ||| g) t x = f t x `choice` g t x

choice :: [a] -> [a] -> [a]
choice [] as =as
choice as _  =as 

-- left-biased choice with FULL backtracking
(|+|) :: Rule -> Rule -> Rule
(f |+| g) t x = f t x `mplus` g t x
-}
-- | Perform a monad under an extended context.
lcl :: MonadState Context m => Context -> m a -> m a
lcl cxt_delta m = do
  cxt <- get
  modify (. cxt_delta)
  result <- m
  put cxt
  return result

-- | Dynamic constructors, used for specifying context extensions.
dinv :: DynR -> DynR
dinv (DynR (Rel a b) r) = DynR (Rel b a) (Rinv r)

dcompl :: DynR -> DynR
dcompl (DynR (Rel a b) r) = DynR (Rel a b) (Rnot r)

dcomp :: DynR -> DynR -> DynR
dcomp (DynR (Rel b c) x) (DynR (Rel a b') y) = fromJust $ do
  Eq <- teq b b'
  return (DynR (Rel a c) (Rcomp b x y))

dker :: DynR -> DynR
dker (DynR (Rel a b) r) = DynR (Rel a a) (Rker b r)

dimg :: DynR -> DynR
dimg (DynR (Rel a b) r) = DynR (Rel b b) (Rimg a r)

ddom :: DynR -> DynR
ddom (DynR (Rel a b) r) = DynR (Rel a a) (Rdom b r)

drng :: DynR -> DynR
drng (DynR (Rel a b) r) = DynR (Rel b b) (Rrng a r)

dincl :: DynR -> DynR -> DynR
dincl (DynR a x) (DynR b y) = fromJust $ do
  Eq <- teq a b
  return (DynR Bool (Rincl a x y))
  
dand :: DynR -> DynR -> DynR
dand (DynR Bool x) (DynR Bool y) = fromJust $ do
  return (DynR Bool (Rand x y))
  
dunion :: DynR -> DynR -> DynR
dunion (DynR a@(Rel _ _) x) (DynR b y) = fromJust $ do
  Eq <- teq a b
  return (DynR a (Runion x y))
  
dinter :: DynR -> DynR -> DynR
dinter (DynR a@(Rel _ _) x) (DynR b y) = fromJust $ do
  Eq <- teq a b
  return (DynR a (Rinter x y))
  
-- | Perform argument rule on all immediate subterms.
all :: Rule -> Rule
all r (Rel a b) (Rinv x)          = do x' <- lcl dinv $ r (Rel b a) x; return (Rinv x')
all r (Rel a b) (Rnot x)        = do x' <- lcl dcompl $ r (Rel a b) x; return (Rnot x')
all r (Rel a c) (Rcomp b x y)     = do 
   x' <- lcl (flip dcomp (DynR (Rel a b) y)) $ r (Rel b c) x; 
   y' <- lcl (dcomp (DynR (Rel b c) x')) $ r (Rel a b) y; 
   return (Rcomp b x' y')
all r (Rel a _) (Rker b x)     = do x' <- lcl dker $ r (Rel a b) x; return (Rker b x')
all r (Rel b _) (Rimg a x)     = do x' <- lcl dimg $ r (Rel a b) x; return (Rimg a x')
all r (Rel a _) (Rdom b x)     = do x' <- lcl ddom $ r (Rel a b) x; return (Rdom b x')
all r (Rel b _) (Rrng a x)     = do x' <- lcl drng $ r (Rel a b) x; return (Rrng a x')
all r t (Runion x y)           = do 
   x' <- lcl (flip dunion (DynR t y)) $ r t x
   y' <- lcl (dunion (DynR t x')) $ r t y
   return (Runion x' y')
all r t (Rinter x y)           = do 
   x' <- lcl (flip dinter (DynR t y)) $ r t x
   y' <- lcl (dinter (DynR t x')) $ r t y
   return (Rinter x' y')
{- Still needs context updating ...
all r (Rel c a) (Runder b x y)   = do 
   x' <- r (Rel a b) x; 
   y' <- r (Rel c b) y; 
   return (Runder b x' y')
all r (Rel a c) (Rover b x y)           = do 
   x' <- r (Rel b c) x; 
   y' <- r (Rel b a) y; 
   return (Rover b x' y')
-}
all r (Bool) (Rincl f x y)    = do 
  x' <- lcl (flip dincl (DynR f y)) $ r f x
  y' <- lcl (dincl (DynR f x')) $ r f y
  return (Rincl f x' y')
all r (Bool) (Rand x y)    = do 
  x' <- lcl (flip dand (DynR Bool y)) $ r Bool x
  y' <- lcl (dand (DynR Bool x')) $ r Bool y
  return (Rand x' y')
all r t x                  = return x

-- | Perform argument rule on exactly one immediate subterm.
one :: Rule -> Rule
one r (Rel a b) (Rinv x)        = do x' <- lcl dinv $ r (Rel b a) x; return (Rinv x')
one r (Rel a b) (Rnot x)      = do x' <- lcl dcompl $ r (Rel a b) x; return (Rnot x')
one r (Rel a c) (Rcomp b x y)   = do 
   x' <- lcl (flip dcomp (DynR (Rel a b) y)) $ r (Rel b c) x; return (Rcomp b x' y)
   `mplus` 
   do y' <- lcl (dcomp (DynR (Rel b c) x)) $ r (Rel a b) y; return (Rcomp b x y')
one r (Rel a _) (Rker b x)      = do x' <- lcl dker $ r (Rel a b) x; return (Rker b x')
one r (Rel b _) (Rimg a x)      = do x' <- lcl dimg $ r (Rel a b) x; return (Rimg a x')
one r (Rel a _) (Rdom b x)      = do x' <- lcl ddom $ r (Rel a b) x; return (Rdom b x')
one r (Rel b _) (Rrng a x)      = do x' <- lcl drng $ r (Rel a b) x; return (Rrng a x')
{- Still needs context updating ...
one r t (Rinter x y)            = do x' <- r t x; return (Rinter x' y)
                                  `mplus`
                                  do y' <- r t y; return (Rinter x y')
one r t (Runion x y)            = do x' <- r t x; return (Runion x' y)
                                  `mplus`
                                  do y' <- r t y; return (Runion x y')
one r (Rel c a) (Runder b x y)  = do 
   x' <- r (Rel a b) x; return (Runder b x' y)
   `mplus`
   do y' <- r (Rel c b) y; return (Runder b x y')
one r (Rel a c) (Rover b x y)   = do 
   x' <- r (Rel b c) x; return (Rover b x' y)
   `mplus`
   do y' <- r (Rel b a) y; return (Rover b x y')                            
-}
one r (Bool) (Rincl f x y)     = do x' <- lcl (flip dincl (DynR f y)) $ r f x; return (Rincl f x' y)
                                    `mplus`
                                  do y' <- lcl (dincl (DynR f x)) $ r f y; return (Rincl f x y')
one r (Bool) (Rand x y)    =  
  do x' <- lcl (flip dand (DynR Bool y)) $ r Bool x; return (Rand x' y)
  `mplus`
  do y' <- lcl (dand (DynR Bool x)) $ r Bool y; return (Rand x y')
one r t x                       = mzero


--------------------------------------------------------------------------------
-- STRATEGIC LIBRARY
--------------------------------------------------------------------------------

try :: Rule -> Rule
try s = s ||| nop

many :: Rule -> Rule
many r = try (r >>> many r)

-- | Same as many
repeats :: Rule -> Rule
repeats s = try (s >>> repeats s)

full_td :: Rule -> Rule
full_td s = s >>> all (full_td s)

full_bu :: Rule -> Rule
full_bu s = all (full_bu s) >>> s

once_td :: Rule  -> Rule      
once_td s = s ||| one (once_td s)

once_bu :: Rule  -> Rule      
once_bu s = one (once_bu s) ||| s 

stop_td :: Rule  -> Rule  
stop_td s = s ||| all(stop_td s)

path_td :: Rule  -> Rule  
path_td s = s >>> (all (nop) ||| one (path_td s))

innermost :: Rule -> Rule
innermost s = all (innermost s) >>> try (s >>> innermost s)

beloweq :: Rule -> Rule -> Rule
beloweq s1 s2 = once_td (s2 >>> (once_td s1))    

untileq :: Rule -> Rule -> Rule
untileq s1 s2 = s1 >>> ( s2 >>> one (untileq s1 s2))


--------------------------------------------------------------------------------
-- STRATEGIES FOR RELATIONAL ALGEBRA
--------------------------------------------------------------------------------

-- | Bring relational expression into normal form where
--   composition is associated to the left, img and ker
--   do not occur, and converse has been distributed over 
--   composition, etc.
normalize :: Rule
normalize = innermost (comp_assocl ||| normalize1)

-- previous normalize 
normalizea :: Rule
normalizea = innermost (innermost comp_assocl >>> normalize1)


normalize1 :: Rule
normalize1  =  desugar ||| 
               simplify1 ||| simplify2 ||| simplify3 |||
               elim1 ||| elim2 ||| elim3 |||
               truth
  where
    desugar     = ker_def ||| img_def
    simplify1   = inv_empty ||| inv_comp ||| inv_inv ||| inv_id  ||| comp_id |||
                  comp_empty ||| neg_comp ||| neg_id ||| dom_inv ||| rng_inv |||
                  union_contravariance ||| union_fusion ||| union_split |||
                  inter_split
    simplify2   = corefl_symm ||| corefl_trans
    simplify3   = dom_id_entire ||| const_fusion ||| bang_fusion
    elim1       = img_elim ||| ker_elim_constant ||| ker_elim 
    elim2       = dom_elim ||| rng_elim 
    elim3       = not_dom_cancel ||| dom_cancel ||| dom_comp_cancel
    
    truth        = incl_refl ||| incl_bottom ||| incl_top

{-
-- 3st CHANGE
simplify :: Rule
simplify = normalize >>> all_and simplify1 >>> innermost and_true
    
simplify1 :: Rule
simplify1 = ( (shunt >>> simplify)
              >>>
              (qed ||| prove_stronger ||| nop)
            )
            |+|
            prove_stronger1
            |||
            nop
-}


-- | Attempt to reduce the given proposition to a tautology.
--   Applies normalization rules together with proof steps.
--   Shunting and strengthening rules are only applied as last resorts.
prove :: Rule
prove _ x@(Rand p q) = trace "HELLPPP" $ do
  trace "HELP" (return ())
  p' <- lcl (flip dand (DynR Bool q)) $ prove Bool p
  q' <- lcl (dand (DynR Bool p')) $ prove Bool q
  x' <- print_msg "and -> " typeof x
  innermost and_true Bool (Rand p' q')
prove t x = do
  x' <- print_msg "PROVE" t x
  prove' t x'

prove' :: Rule
prove' = innermost step >>> try (qed ||| (print_msg "LAST RESORT" >>> last_resort))
  where
    step       = truth ||| and_true ||| comp_assocl ||| normalize1
    truth      = incl_refl ||| incl_bottom ||| incl_bottom
    
    last_resort  = attempts [shuntc,shunt,strengthen]
    shunt      = gc_shunt ||| gc_RSshunt 
    shuntc     = gc_shuntc ||| (once_td gc_RSshuntc)
    strengthen = once_td corefl_cancel'

{-
prove' :: Rule
prove' = innermost step >>> (step ||| last_resort)
  where
    step       = truth ||| and_true ||| comp_assocl ||| normalize1
    truth      = incl_refl ||| incl_bottom
    
    last_resort  = shuntc |+| shunt |+| strengthen
    shunt      = gc_shunt |+| gc_RSshunt 
    shuntc     = gc_shuntc |+| gc_RSshuntc
    strengthen = corefl_cancel'
-}

-- DOES NOT TERMINATE!
prove'' :: Rule
prove'' = innermost step >>> try (qed ||| (once_td proof_step >>> prove''))
  where
    step         = truth ||| and_true ||| comp_assocl ||| normalize1
    truth        = incl_refl ||| incl_bottom ||| incl_top

    proof_step   = (shuntc |+| shunt) ||| strengthen
    shunt        = gc_shunt |+| gc_RSshunt 
    shuntc       = gc_shuntc |+| gc_RSshuntc
    strengthen   = corefl_cancel' |+| inter_fusion
    
simplify :: Rule
simplify = normalize >>> all_and process_conjunct >>> innermost and_true
  where
    process_conjunct = (shunt_conjunct |+| strengthen_conjunct ) ||| nop
    shunt_conjunct   = shunt >>> simplify 
    strengthen_conjunct = strengthen >>> simplify >>> qed
    strengthen   = corefl_cancel' |+| inter_fusion
    
shunt :: Rule
shunt   = shuntc |+| shunt
  where
    shunt        = gc_shunt ||| gc_RSshunt' 
    shuntc       = gc_shuntc ||| gc_RSshuntc
    
prove_stronger :: Rule
prove_stronger = normalize >>> all_and prove_stronger1 >>> innermost and_true

prove_stronger1 :: Rule
prove_stronger1 = strengthen >>> simplify >>> qed
  where
    strengthen   = corefl_cancel'
    

-- | Check whether truth has been established.
qed :: Rule
qed _ e@(Rvar "True" _)  = return e
qed _ _                  = mzero

-- | Apply argument rule, but backtrack from it if it does not lead to truth.
attempt :: Rule -> Rule
attempt r = print_msg "ATTEMPT" >>> r >>> prove >>> print_msg "QED" >>> qed

-- | Attempt application of each argument rule, until one leads to truth.
attempts :: [Rule] -> Rule
attempts []      = \_ _ -> tell [(error "_|_","Botched","_|_")] >> mzero
attempts (r:rs)  = attempt r ||| attempts rs

print_trace :: Rule -> Rule
print_trace r t x = do
  (x',w) <- listen (r t x)
  trace ("ATTEMPT\n"++show w++"\nFAILED") (return x')
  
print_msg :: String -> Rule
print_msg msg t x = trace (msg++" on "++show x) $ return x


-- | Introduce syntactic sugar.
sugar :: Rule
sugar = innermost (ker_def' ||| img_def')

--------------------------------------------------------------------------------
-- auxs. functions

-- First projection
pi1 :: R (Rel (a,b) a)
pi1 = Rvar "pi1" [Entire,Simple,Surjective]

-- Second projection
pi2 :: R (Rel (a,b) b)
pi2 = Rvar "pi2" [Entire,Simple,Surjective]

-- Constant function
k    :: Show b => b -> R (Rel a b)
k b  = Rvar ("const "++show b) [Entire,Simple]

kk      :: Type b -> String -> R (Rel a b)
kk _ b  = Rvar ("const "++b) [Entire,Simple]

-- Testing for constant function
isConstant :: R (Rel a b) -> Bool
isConstant (Rvar n _) 
  | "const " `isPrefixOf` n
  = True
isConstant _ = False

rsplit :: Type (Rel a (b,c)) 
       -> R (Rel a b) -> R (Rel a c) -> R (Rel a (b, c))
rsplit (Rel _ (Prod b c)) r s = Rinter (Rcomp b (Rinv pi1) r) (Rcomp c (Rinv pi2) s)
   
--------------------------------------------------------------------------------
-- TESTS AND EXAMPLES
--------------------------------------------------------------------------------
-- Buyer-Seller Model
type User_id = Int
type Item_id = Int
type Description = String
type Name = String
type Price = Float
type Balance = Float

users :: R (Rel User_id (Name,Balance))
users = Rvar "users" [Simple]

items:: R (Rel Item_id Description)
items = Rvar "items" [Simple]

forSale:: R (Rel Item_id (User_id,Price))
forSale = Rvar "forSale" [Simple]

wanted:: R (Rel (Item_id,User_id) Price)
wanted = Rvar "wanted" [Simple] 

deal:: R (Rel Item_id (User_id, Price))
deal = Rvar "deal" [Simple]

-- Selection on domain with simple+injective predicate gives map
balanceGreatZero :: R (Rel (Name,Balance) Balance)
balanceGreatZero = Rvar "balance>0" [Simple,Injective] 

selectUsersbalanceGreatZero ::  R (Rel User_id Balance)
selectUsersbalanceGreatZero = balanceGreatZero .*. users .*. Rid 

myProofObligation :: R Bool
myProofObligation = propSimple typeof selectUsersbalanceGreatZero

rsnd :: R (Rel (a,b) b)
rsnd = Rvar "snd" [Entire,Simple,Surjective]

--"Select all Items for which at least one user want it"
-- Prove to be simple? ...OK
wantedItems :: R (Rel Item_id ())
wantedItems = bang .*. wanted .*. Rinv pi1

-- Prove that roverride preserves simplicity (generic, with variables)? ...OK
over2TablesUsers :: R (Rel User_id (Name,Balance))
over2TablesUsers = roverride typeof m n
  where
    m :: R (Rel User_id (Name,Balance))
    m = Rvar "m" [Simple]
    n :: R (Rel User_id (Name,Balance))
    n = Rvar "n" [Simple]

-- roverride preserves simplicity (pre-split, paper 3rd example)...
-- 
ex12' :: R(Rel Item_id User_id) -> 
         R(Rel Item_id Price)   -> 
         R (Rel Item_id (User_id,Price))
ex12' u x = roverride typeof m n
  where
    m :: R (Rel Item_id (User_id, Price))
    m = forSale
    n ::R (Rel Item_id (User_id,Price))
    n = withUser u x
    withUser :: R(Rel Item_id User_id) -> R(Rel Item_id Price) -> 
                R(Rel Item_id (User_id,Price))
    withUser u1 m1 = Rvar "withUser" [Simple]  
--    withUser u1 m1 = undefined --Img((*** Exception: Prelude.undefined
--    withUser u1 m1 = rsplit typeof u1 m1  -- Not produces True (idem put...)

-- to prove ex12'
paru :: R(Rel Item_id User_id)
paru =  (kk typeof "u")
parx:: R (Rel Item_id Price)
parx = Rvar "Map Item Price" [Simple]
   
-- rupdate preserves simplicity? OK
updateUsers = rupdate typeof g phi m
  where
    m :: R (Rel User_id (Name,Balance))
    m = Rvar "m" [Simple]
    g = Rvar "g" [Simple,Entire] :: R (Rel (Name,Balance) (Name,Balance))
    phi :: R (Rel User_id User_id )
    phi = Rvar "phi" [Injective,Simple,Coreflexive] 

listWantedItems :: R (Rel Item_id Price)
listWantedItems = wanted .*. Rinv pi1

-- paper, pag.10
mbs012 = Rinv (n .*. Rinv (Rnot (Rdom typeof n)) .*. Rinv m) 
  where
    m :: R (Rel Item_id (User_id,Price))
    m = Rvar "M" [Simple]
    n :: R (Rel Item_id (User_id,Price))
    n = Rvar "N" [Simple]
    
-- all OK (True)
-- to Prove const_fusion
mbs0 =  propSimple typeof (c .*. forSale) 
  where
    c = kk One "u"

-- ***************************************************************************
-- Movie database (IMDB)

type Year = Int
type Title = String
type Director = String

movies :: R (Rel (Year,Title) Director)
movies = Rvar "movies" [Simple]

yearEquals2004 :: R (Rel Year Year)
yearEquals2004 = Rvar "year==2004" [Simple,Injective]  -- ??

yearEquals2004split :: R (Rel (Year,Title) (Year,Title))
yearEquals2004split = Rvar "g" [Simple,Injective]
-- yearEquals2004split = Rvar "(year==2004 >< id)" [Simple,Injective]

select2004movies :: R (Rel Title Director)
select2004movies = Rid .*. movies .*. Rinv (rsnd .*. yearEquals2004split)

select2004movies' :: R (Rel (Year ,Title)  Director)
select2004movies' = Rid .*. movies .*. Rinv (yearEquals2004split)

myProofObligationa :: R Bool
myProofObligationa = propSimple typeof select2004movies

-- Two ways of writing "Select all years for which at least one movie exists".
-- Prove to be simple? Ok.
ex8 :: R (Rel Year ())
ex8 = bang .*. movies .*. Rinv pi1

-- Prove to be coreflexive? OK.
ex9 :: R (Rel Year Year)
ex9 = pi1 .*. Rinv (Rdom typeof movies) .*. (Rdom typeof movies) .*. Rinv pi1

ex11 :: R (Rel (Year,Title) Director)
ex11 = Runion m n
  where
    m :: R (Rel (Year,Title) Director)
    m = Rvar "m" [Simple]
    n :: R (Rel (Year,Title) Director)
    n = m

-- Prove that roverride preserves simplicity. OK.
ex12 :: R (Rel (Year,Title) Director)
ex12 = roverride typeof m n
  where
    m :: R (Rel (Year,Title) Director)
    m = Rvar "m" [Simple]
    n :: R (Rel (Year,Title) Director)
    n = Rvar "n" [Simple]


-- Prove to be coreflexive? OK
ex13 :: R (Rel Director Director)
ex13 = Rcomp typeof n (Rcomp typeof (Rinv (Rnot (Rdom typeof n))) (Rinv m))
  where
    m :: R (Rel (Year,Title) Director)
    m = Rvar "m" [Simple]
    n :: R (Rel (Year,Title) Director)
    n = Rvar "n" [Simple]

ex14 :: R (Rel Director Director)
ex14 = Rcomp typeof m (Rcomp typeof (Rnot (Rdom typeof n)) (Rinv n))
  where
    m :: R (Rel (Year,Title) Director)
    m = Rvar "m" [Simple]
    n :: R (Rel (Year,Title) Director)
    n = Rvar "n" [Simple]

-- Coreflexive ? OK
ex15 :: R (Rel Director Director)
ex15 = Rcomp typeof m (Rcomp typeof x (Rcomp typeof (Rinv x) (Rinv m)))
  where
    m :: R (Rel (Year,Title) Director)
    m = Rvar "m" [Simple]
    n :: R (Rel (Year,Title) Director)
    n = Rvar "n" [Simple]
    x = Rnot (Rdom typeof n)

-- rupdate preserves simplicity?.....
ex16 = rupdate typeof g phi m
  where
    m :: R (Rel (Year,Title) Director)
    m = Rvar "m" [Simple]
    g = Rvar "g" [Simple,Entire] :: R (Rel Director Director)
    phi :: R (Rel (Year,Title) (Year,Title))
    phi = Rvar "phi" [Injective,Simple,Coreflexive] 
    
