{-# OPTIONS -fglasgow-exts #-}
{-# OPTIONS -fallow-undecidable-instances #-}
{-# OPTIONS -fallow-overlapping-instances #-}
-----------------------------------------------------------------------------
-- |
-- Maintainer  :  joost.visser@di.uminho.pt, alcino@di.uminho.pt
-- Stability   :  experimental
-- Portability :  portable
-----------------------------------------------------------------------------

module Transform.Relation.Type where

--import Data.Relation.SetOfPairs as Rel
import Data.Set as Set
import Data.Map as Map
import Data.List as List
import Control.Monad

-----------------------------------------------------------------------------
-- * Representation of Types

--data Includes a = Includes a a

data Rel a b

-- | A type-safe representation of types.
data Type a where

    -- Without fixpoints
    Int :: Type Int
    Bool :: Type Bool
    Char :: Type Char
    String :: Type String
    Float :: Type Float
    One :: Type One
    List :: Type a -> Type [a]
    Maybe :: Type a -> Type (Maybe a) 
    Set :: Type a -> Type (Set a)
    Map :: Type a -> Type b -> Type (Map a b)
    Either :: Type a -> Type b -> Type (Either a b)
    Prod :: Type a -> Type b -> Type (a,b)
    Func :: Type a -> Type b -> Type (a -> b)
    Tag :: String -> Type a -> Type a
    Rel :: Type a -> Type b -> Type (Rel a b)
--    Incl :: Type x -> Type (Includes x)
    
    -- used to represent XML elements
    EL :: String -> Type a -> Type (El a)
    Tuple :: Tuple a -> Type a
    
    Dynamic :: Type Dynamic

data Tuple a where
    -- TNil :: Tuple ()
    TCons :: Type a -> Tuple b -> Tuple (a,b)
    -- TAppend :: Tuple a -> Tuple b -> Tuple (a,b)
    TSing :: Type a -> Tuple a

-- | Representation of XML elements.
data El a = El a 
  deriving (Eq,Show)

-- | A type-safe representation of lifted functors.
data Fctr f where
    Id    :: Fctr Id
    K     :: Type a -> Fctr (K a)
    (:+:) :: Fctr g -> Fctr h -> Fctr (g:+:h)
    (:*:) :: Fctr g -> Fctr h -> Fctr (g:*:h)
    (:@:) :: Functor g => (forall a . Type a -> Type (g a)) -> Fctr h -> Fctr (g:@:h)

instance Show (Type a) where
    show Int = "Int"
    show Bool = "Bool"
    show Char = "Char"
    show String = "String"
    show Float = "Float"
    show One = "One"
    show (List x) = "(List "++(show x)++")"
    show (Set x) = "(Set "++(show x)++")"
    show (Map x y) = "(Map "++(show x)++" "++(show y)++")"
    show (Either x y) = "(Either "++(show x)++" "++(show y)++")"
    show (Prod x y) = "(Prod "++(show x)++" "++(show y)++")"
    show (Func x y) = "("++show x++" -> "++show y++")"
    show (Tag n t) = "(Tag "++(show n)++" "++(show t)++")"
    show (EL l a)       = "(EL"++show l++" "++show a++")"
    show (Rel x y) = "(Rel "++(show x)++" "++(show y)++")"
--    show (Incl x) = "(Incl "++(show x)++")"

    show (Dynamic) = "Dynamic"
    
instance Show (Fctr f) where
  show Id = "Id"
  show (K t) = "(K "++show t++")"
  show (f:*:g) = "("++show f++":*:"++show g++")"
  show (f:+:g) = "("++show f++":+:"++show g++")"
  show (f:@:g) = "(<?>:@:"++show g++")"
 
----------------------------------------
-- ** Type One

-- data One
type One = ()

_L :: One
_L = ()

-- instance Show One where
--     show _ = "_L"

----------------------------------------
-- ** Type Dynamic

-- | Dynamic types
data Dynamic where
    Dyn :: Type a -> a -> Dynamic

-----------------------------------------------------------------------------
-- * Deriving type representations from types.

-- | Typeable class of types for which we can derive representations.
class Typeable a where
    typeof :: Type a

instance Typeable Int where
    typeof = Int

instance Typeable Float where
    typeof = Float

instance Typeable () where
    typeof = One

instance Typeable Bool where
    typeof = Bool

instance Typeable Char where
    typeof = Char

instance (Typeable a, Typeable b) => Typeable (a,b) where
    typeof = Prod typeof typeof

instance (Typeable a, Typeable b) => Typeable (Either a b) where
    typeof = Either typeof typeof

instance (Typeable a, Typeable b) => Typeable (Map a b) where
    typeof = Map typeof typeof

instance (Typeable a, Typeable b) => Typeable (a -> b) where
    typeof = Func typeof typeof
    
instance (Typeable a, Typeable b) => Typeable (Rel a  b) where
    typeof = Rel typeof typeof   

--instance (Typeable a) => Typeable (Includes a) where
--    typeof = Incl typeof   

instance Typeable Dynamic where
    typeof = Dynamic

instance Typeable String where
    typeof = String

instance Typeable a => Typeable [a] where
    typeof = List typeof

instance Typeable a => Typeable (Set a) where
    typeof = Set typeof

-----------------------------------------------------------------------------
-- * Type-indexed functions

----------------------------------------
-- ** Some generic functions

-- | Explicitly contrain a value to a particular type.
constrain :: Type a -> a -> a
constrain _ a = a

-- | Explicitly coerce a value of a given type to another given type.
coerce :: MonadPlus m => Type a -> Type b -> a -> m b
coerce a b x = do Eq <- teq a b
                  return x

----------------------------------------
-- ** Equality of types

-- | Data type to express type-equality.    
data Equal a b where
    Eq :: Equal a a
    
-- | Test equality of types.
--   If the types are equal, a witness 'Eq' of the equality is returned.
teq :: MonadPlus m => Type a -> Type b -> m (Equal a b)
teq Dynamic Dynamic = return Eq
teq Int Int = return Eq
teq Float Float = return Eq
teq Bool Bool = return Eq
teq Char Char = return Eq
teq One One = return Eq
teq String String = return Eq
teq (List a) (List b) = 
    do Eq <- teq a b
       return Eq
teq (Set a) (Set b) = 
    do Eq <- teq a b
       return Eq
teq (Maybe a) (Maybe b) =
    do Eq <- teq a b
       return Eq
teq (Map a b) (Map c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (Prod a b) (Prod c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (Either a b) (Either c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (Func a b) (Func c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
teq (Rel a b) (Rel c d) =
    do Eq <- teq a c
       Eq <- teq b d
       return Eq
--teq (Incl a) (Incl c) =
--    do Eq <- teq a c
--       return Eq
teq (Tag _ a) b =
    do Eq <- teq a b
       return Eq
teq a (Tag _ b) =
    do Eq <- teq a b
       return Eq
teq (EL l a) (EL l' a') = do
  guard (l==l')
  Eq <- teq a a'
  return Eq
teq t t' 
   | show t == show t'
   = error $ "Should be equal: "++show t++" and "++show t'
teq _ _ = mzero
  
teq' :: Type a -> Type b -> Maybe (Equal a b)
teq' = teq

-----------------------------------------------------------------------------
-- * Representation of recursive datatypes.

----------------------------------------
-- ** Standard functors

infixr 5 :+:
infixr 6 :*:
infixr 9 :@:

newtype Id a        = Ident {unIdent :: a}       deriving Eq
newtype K b a       = Const {unConst :: b}       deriving Eq
data (g :+: h) a    = Inl (g a) | Inr (h a)      deriving Eq
data (g :*: h) a    = Pair (g a) (h a)           deriving Eq
newtype (g :@: h) a = Comp {unComp :: g (h a)}   deriving Eq

instance Functor Id where
    fmap f = Ident . f . unIdent

instance Functor (K a) where
    fmap f = Const . id . unConst

instance (Functor f, Functor g) => Functor (f :+: g) where
    fmap f (Inl x) = Inl (fmap f x)
    fmap f (Inr x) = Inr (fmap f x)

instance (Functor f, Functor g) => Functor (f :*: g) where
    fmap f (Pair x y) = Pair (fmap f x) (fmap f y)

instance (Functor g, Functor h) => Functor (g :@: h)
    where fmap f = Comp . fmap (fmap f) . unComp

{-
----------------------------------------
-- ** Fixpoints and recursion patterns

-- | Explicit fixpoint operator
newtype Mu f = In {out :: f (Mu f)}

-- This instance requires -fallow-undecidable-instances.
instance Eq (f (Mu f)) => Eq (Mu f) where
  (In fmf) == (In gmg) = fmf==gmg

-- | Catamorphism or fold
cata :: Functor f => (f a -> a) -> Mu f -> a
cata g = g . fmap (cata g) . out

-- | Anamorphism or unfold
ana :: Functor f => (a -> f a) -> a -> Mu f
ana g = In . fmap (ana g) . g

-- | Natural transformation
nu :: Functor g => (forall a. f a -> g a) -> Mu f -> Mu g
nu t = In . fmap (nu t) . t . out

----------------------------------------
-- ** Functors and monads

-- | Class of functors that distribute over monads.
class Functor f => Dist f where
    dist :: Monad m => f (m a) -> m (f a)

instance Dist Id where
    dist (Ident x) = do y <- x
			return (Ident y)

instance Dist (K a) where
    dist (Const x) = return (Const x)

instance (Dist f, Dist g) => Dist (f :+: g) where
    dist (Inl x) = do y <- dist x
		      return (Inl y)
    dist (Inr x) = do y <- dist x
		      return (Inr y)

instance (Dist f, Dist g) => Dist (f :*: g) where
    dist (Pair l r) = do x <- dist l
			 y <- dist r
			 return (Pair x y)

instance (Dist g, Dist h) => Dist (g :@: h) where
    dist (Comp x) = do y <- return $ fmap dist x
		       z <- dist y
		       return (Comp z)

instance Dist [] where
    dist = sequence


-- | Monadic catamorphism.
mcata :: (Monad m, Dist f) => (f a -> m a) -> Mu f -> m a
mcata g = cata (\x -> dist x >>= g)
-}

-----------------------------------------------------------------------------
-- * Strategy combinators on type representations

-- | One-step type-unifying traversal combinator:
allType :: (forall a . Type a -> [b]) -> (forall a . Type a -> [b])
allType f Int                   = []
allType f Float                 = []
allType f String                = []
allType f (Prod a b)            = f a ++ f b
allType f (Either a b)          = f a ++ f b
allType f (List a)              = f a
allType f (EL l a)              = f a

-- | Type-unifying complete traversal.
crushType :: (forall a . Type a -> [b]) -> (forall a . Type a -> [b])
crushType f a = f a ++ allType (crushType f) a

-- | Type-unifying traversal that does not go below successful nodes.
crushTypeStop :: (forall a . Type a -> [b]) -> (forall a . Type a -> [b])
crushTypeStop f a = if List.null atRoot then belowRoot else atRoot
  where
    atRoot = f a
    belowRoot = allType (crushTypeStop f) a

